@@ -9,6 +9,7 @@ import (
 	"bytes"
 	"encoding/binary"
 	"encoding/hex"
+	"fmt"
 	"io"
 	"io/ioutil"
 	"log"
@@ -18,6 +19,7 @@ import (
 	"reflect"
 	"runtime"
 	"strings"
+	"sync"
 	"testing"
 	"time"
 
@@ -85,6 +87,7 @@ func TestNewReaderMismatch(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
+	defer dec.Close()
 	var tmp [8]byte
 	xx := xxhash.New()
 	var cHash int
@@ -94,7 +97,7 @@ func TestNewReaderMismatch(t *testing.T) {
 		n, err := io.ReadFull(dec, buf)
 		if err != nil {
 			if err != io.EOF && err != io.ErrUnexpectedEOF {
-				t.Fatal(err)
+				t.Fatal("block", cHash, "err:", err)
 			}
 		}
 		if n > 0 {
@@ -146,6 +149,68 @@ func TestNewDecoder(t *testing.T) {
 	testDecoderDecodeAll(t, "testdata/decoder.zip", dec)
 }
 
+func TestNewDecoderMemory(t *testing.T) {
+	defer timeout(60 * time.Second)()
+	var testdata bytes.Buffer
+	enc, err := NewWriter(&testdata, WithWindowSize(64<<10), WithSingleSegment(false))
+	if err != nil {
+		t.Fatal(err)
+	}
+	// Write 256KB
+	for i := 0; i < 256; i++ {
+		tmp := strings.Repeat(string([]byte{byte(i)}), 1024)
+		_, err := enc.Write([]byte(tmp))
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+	err = enc.Close()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	var n = 5000
+	if testing.Short() {
+		n = 200
+	}
+
+	var before, after runtime.MemStats
+	runtime.GC()
+	runtime.ReadMemStats(&before)
+
+	var decs = make([]*Decoder, n)
+	for i := range decs {
+		// Wrap in NopCloser to avoid shortcut.
+		input := ioutil.NopCloser(bytes.NewBuffer(testdata.Bytes()))
+		decs[i], err = NewReader(input, WithDecoderConcurrency(1), WithDecoderLowmem(true))
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+
+	// 32K buffer
+	var tmp [128 << 10]byte
+	for i := range decs {
+		_, err := io.ReadFull(decs[i], tmp[:])
+		if err != nil {
+			t.Fatal(err)
+		}
+	}
+
+	runtime.GC()
+	runtime.ReadMemStats(&after)
+	size := (after.HeapInuse - before.HeapInuse) / uint64(n) / 1024
+	t.Log(size, "KiB per decoder")
+	// This is not exact science, but fail if we suddenly get more than 2x what we expect.
+	if size > 221*2 && !testing.Short() {
+		t.Errorf("expected < 221KB per decoder, got %d", size)
+	}
+
+	for _, dec := range decs {
+		dec.Close()
+	}
+}
+
 func TestNewDecoderGood(t *testing.T) {
 	defer timeout(30 * time.Second)()
 	testDecoderFile(t, "testdata/good.zip")
@@ -179,6 +244,7 @@ func TestNewReaderRead(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
+	defer dec.Close()
 	_, err = dec.Read([]byte{0})
 	if err == nil {
 		t.Fatal("Wanted error on uninitialized read, got nil")
@@ -263,8 +329,85 @@ func TestNewDecoderSmallFile(t *testing.T) {
 	t.Logf("Decoded %d bytes with %f.2 MB/s", n, mbpersec)
 }
 
+type readAndBlock struct {
+	buf     []byte
+	unblock chan struct{}
+}
+
+func (r *readAndBlock) Read(p []byte) (int, error) {
+	n := copy(p, r.buf)
+	if n == 0 {
+		<-r.unblock
+		return 0, io.EOF
+	}
+	r.buf = r.buf[n:]
+	return n, nil
+}
+
+func TestNewDecoderFlushed(t *testing.T) {
+	if testing.Short() {
+		t.SkipNow()
+	}
+	file := "testdata/z000028.zst"
+	payload, err := ioutil.ReadFile(file)
+	if err != nil {
+		t.Fatal(err)
+	}
+	payload = append(payload, payload...) //2x
+	payload = append(payload, payload...) //4x
+	payload = append(payload, payload...) //8x
+	rng := rand.New(rand.NewSource(0x1337))
+	runs := 100
+	if testing.Short() {
+		runs = 5
+	}
+	enc, err := NewWriter(nil, WithWindowSize(128<<10))
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer enc.Close()
+	for i := 0; i < runs; i++ {
+		wantSize := rng.Intn(len(payload)-1) + 1
+		t.Run(fmt.Sprint("size-", wantSize), func(t *testing.T) {
+			var encoded bytes.Buffer
+			enc.Reset(&encoded)
+			_, err := enc.Write(payload[:wantSize])
+			if err != nil {
+				t.Fatal(err)
+			}
+			err = enc.Flush()
+			if err != nil {
+				t.Fatal(err)
+			}
+
+			// We must be able to read back up until the flush...
+			r := readAndBlock{
+				buf:     encoded.Bytes(),
+				unblock: make(chan struct{}),
+			}
+			defer timeout(5 * time.Second)()
+			dec, err := NewReader(&r)
+			if err != nil {
+				t.Fatal(err)
+			}
+			defer dec.Close()
+			defer close(r.unblock)
+			readBack := 0
+			dst := make([]byte, 1024)
+			for readBack < wantSize {
+				// Read until we have enough.
+				n, err := dec.Read(dst)
+				if err != nil {
+					t.Fatal(err)
+				}
+				readBack += n
+			}
+		})
+	}
+}
+
 func TestDecoderRegression(t *testing.T) {
-	defer timeout(5 * time.Second)()
+	defer timeout(160 * time.Second)()
 	data, err := ioutil.ReadFile("testdata/regression.zip")
 	if err != nil {
 		t.Fatal(err)
@@ -273,15 +416,14 @@ func TestDecoderRegression(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	dec, err := NewReader(nil, WithDecoderConcurrency(1), WithDecoderLowmem(true), WithDecoderMaxMemory(10<<20))
+	dec, err := NewReader(nil, WithDecoderConcurrency(1), WithDecoderLowmem(true), WithDecoderMaxMemory(1<<20))
 	if err != nil {
 		t.Error(err)
 		return
 	}
 	defer dec.Close()
-
-	for _, tt := range zr.File {
-		if !strings.HasSuffix(t.Name(), "") {
+	for i, tt := range zr.File {
+		if !strings.HasSuffix(tt.Name, "") || (testing.Short() && i > 10) {
 			continue
 		}
 		t.Run("Reader-"+tt.Name, func(t *testing.T) {
@@ -295,8 +437,38 @@ func TestDecoderRegression(t *testing.T) {
 				t.Error(err)
 				return
 			}
-			got, err := ioutil.ReadAll(dec)
-			t.Log("Received:", len(got), err)
+			got, gotErr := ioutil.ReadAll(dec)
+			t.Log("Received:", len(got), gotErr)
+
+			// Check a fresh instance
+			r, err = tt.Open()
+			if err != nil {
+				t.Error(err)
+				return
+			}
+			decL, err := NewReader(r, WithDecoderConcurrency(1), WithDecoderLowmem(true), WithDecoderMaxMemory(1<<20))
+			if err != nil {
+				t.Error(err)
+				return
+			}
+			defer decL.Close()
+			got2, gotErr2 := ioutil.ReadAll(decL)
+			t.Log("Fresh Reader received:", len(got2), gotErr2)
+			if gotErr != gotErr2 {
+				if gotErr != nil && gotErr2 != nil && gotErr.Error() != gotErr2.Error() {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+				if (gotErr == nil) != (gotErr2 == nil) {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+			}
+			if !bytes.Equal(got2, got) {
+				if gotErr != nil {
+					t.Log("Buffer mismatch without Reset")
+				} else {
+					t.Error("Buffer mismatch without Reset")
+				}
+			}
 		})
 		t.Run("DecodeAll-"+tt.Name, func(t *testing.T) {
 			r, err := tt.Open()
@@ -308,8 +480,71 @@ func TestDecoderRegression(t *testing.T) {
 			if err != nil {
 				t.Error(err)
 			}
-			got, err := dec.DecodeAll(in, nil)
-			t.Log("Received:", len(got), err)
+			got, gotErr := dec.DecodeAll(in, nil)
+			t.Log("Received:", len(got), gotErr)
+
+			// Check if we got the same:
+			decL, err := NewReader(nil, WithDecoderConcurrency(1), WithDecoderLowmem(true), WithDecoderMaxMemory(1<<20))
+			if err != nil {
+				t.Error(err)
+				return
+			}
+			defer decL.Close()
+			got2, gotErr2 := decL.DecodeAll(in, nil)
+			t.Log("Fresh Reader received:", len(got2), gotErr2)
+			if gotErr != gotErr2 {
+				if gotErr != nil && gotErr2 != nil && gotErr.Error() != gotErr2.Error() {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+				if (gotErr == nil) != (gotErr2 == nil) {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+			}
+			if !bytes.Equal(got2, got) {
+				if gotErr != nil {
+					t.Log("Buffer mismatch without Reset")
+				} else {
+					t.Error("Buffer mismatch without Reset")
+				}
+			}
+		})
+		t.Run("Match-"+tt.Name, func(t *testing.T) {
+			r, err := tt.Open()
+			if err != nil {
+				t.Error(err)
+				return
+			}
+			in, err := ioutil.ReadAll(r)
+			if err != nil {
+				t.Error(err)
+			}
+			got, gotErr := dec.DecodeAll(in, nil)
+			t.Log("Received:", len(got), gotErr)
+
+			// Check a fresh instance
+			decL, err := NewReader(bytes.NewBuffer(in), WithDecoderConcurrency(1), WithDecoderLowmem(true), WithDecoderMaxMemory(1<<20))
+			if err != nil {
+				t.Error(err)
+				return
+			}
+			defer decL.Close()
+			got2, gotErr2 := ioutil.ReadAll(decL)
+			t.Log("Reader Reader received:", len(got2), gotErr2)
+			if gotErr != gotErr2 {
+				if gotErr != nil && gotErr2 != nil && gotErr.Error() != gotErr2.Error() {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+				if (gotErr == nil) != (gotErr2 == nil) {
+					t.Error(gotErr, "!=", gotErr2)
+				}
+			}
+			if !bytes.Equal(got2, got) {
+				if gotErr != nil {
+					t.Log("Buffer mismatch")
+				} else {
+					t.Error("Buffer mismatch")
+				}
+			}
 		})
 	}
 }
@@ -331,6 +566,7 @@ func TestDecoder_Reset(t *testing.T) {
 	if err != nil {
 		t.Fatal(err)
 	}
+	defer dec.Close()
 	decoded, err := dec.DecodeAll(dst, nil)
 	if err != nil {
 		t.Error(err, len(decoded))
@@ -422,6 +658,9 @@ func TestDecoderMultiFrame(t *testing.T) {
 				t.Fatal(err)
 			}
 			got, err := ioutil.ReadAll(dec)
+			if err != nil {
+				t.Fatal(err)
+			}
 			err = dec.Reset(bytes.NewBuffer(in))
 			if err != nil {
 				t.Fatal(err)
@@ -481,6 +720,9 @@ func TestDecoderMultiFrameReset(t *testing.T) {
 				t.Fatal(err)
 			}
 			got, err := ioutil.ReadAll(dec)
+			if err != nil {
+				t.Fatal(err)
+			}
 			err = dec.Reset(bytes.NewBuffer(in))
 			if err != nil {
 				t.Fatal(err)
@@ -534,8 +776,8 @@ func testDecoderFile(t *testing.T, fn string) {
 		return
 	}
 	defer dec.Close()
-	for _, tt := range zr.File {
-		if !strings.HasSuffix(tt.Name, ".zst") {
+	for i, tt := range zr.File {
+		if !strings.HasSuffix(tt.Name, ".zst") || (testing.Short() && i > 20) {
 			continue
 		}
 		t.Run("Reader-"+tt.Name, func(t *testing.T) {
@@ -622,6 +864,9 @@ func BenchmarkDecoder_DecoderSmall(b *testing.B) {
 				b.Fatal(err)
 			}
 			got, err := ioutil.ReadAll(dec)
+			if err != nil {
+				b.Fatal(err)
+			}
 			b.SetBytes(int64(len(got)))
 			b.ReportAllocs()
 			b.ResetTimer()
@@ -670,6 +915,9 @@ func BenchmarkDecoder_DecodeAll(b *testing.B) {
 				b.Fatal(err)
 			}
 			got, err := dec.DecodeAll(in, nil)
+			if err != nil {
+				b.Fatal(err)
+			}
 			b.SetBytes(int64(len(got)))
 			b.ReportAllocs()
 			b.ResetTimer()
@@ -683,6 +931,56 @@ func BenchmarkDecoder_DecodeAll(b *testing.B) {
 	}
 }
 
+func BenchmarkDecoder_DecodeAllParallel(b *testing.B) {
+	fn := "testdata/benchdecoder.zip"
+	data, err := ioutil.ReadFile(fn)
+	if err != nil {
+		b.Fatal(err)
+	}
+	zr, err := zip.NewReader(bytes.NewReader(data), int64(len(data)))
+	if err != nil {
+		b.Fatal(err)
+	}
+	dec, err := NewReader(nil)
+	if err != nil {
+		b.Fatal(err)
+		return
+	}
+	defer dec.Close()
+	for _, tt := range zr.File {
+		if !strings.HasSuffix(tt.Name, ".zst") {
+			continue
+		}
+		b.Run(tt.Name, func(b *testing.B) {
+			r, err := tt.Open()
+			if err != nil {
+				b.Fatal(err)
+			}
+			defer r.Close()
+			in, err := ioutil.ReadAll(r)
+			if err != nil {
+				b.Fatal(err)
+			}
+			got, err := dec.DecodeAll(in, nil)
+			if err != nil {
+				b.Fatal(err)
+			}
+			b.SetBytes(int64(len(got)))
+			b.ReportAllocs()
+			b.ResetTimer()
+			b.RunParallel(func(pb *testing.PB) {
+				got := make([]byte, len(got))
+				for pb.Next() {
+					_, err = dec.DecodeAll(in, got[:0])
+					if err != nil {
+						b.Fatal(err)
+					}
+				}
+			})
+		})
+	}
+}
+
 /*
 func BenchmarkDecoder_DecodeAllCgo(b *testing.B) {
 	fn := "testdata/benchdecoder.zip"
@@ -717,7 +1015,7 @@ func BenchmarkDecoder_DecodeAllCgo(b *testing.B) {
 			b.ReportAllocs()
 			b.ResetTimer()
 			for i := 0; i < b.N; i++ {
-				got, err = zstd.Decompress(got[:0], in)
+				got, err = zstd.Decompress(got, in)
 				if err != nil {
 					b.Fatal(err)
 				}
@@ -726,6 +1024,50 @@ func BenchmarkDecoder_DecodeAllCgo(b *testing.B) {
 	}
 }
 
+func BenchmarkDecoder_DecodeAllParallelCgo(b *testing.B) {
+	fn := "testdata/benchdecoder.zip"
+	data, err := ioutil.ReadFile(fn)
+	if err != nil {
+		b.Fatal(err)
+	}
+	zr, err := zip.NewReader(bytes.NewReader(data), int64(len(data)))
+	if err != nil {
+		b.Fatal(err)
+	}
+	for _, tt := range zr.File {
+		if !strings.HasSuffix(tt.Name, ".zst") {
+			continue
+		}
+		b.Run(tt.Name, func(b *testing.B) {
+			r, err := tt.Open()
+			if err != nil {
+				b.Fatal(err)
+			}
+			defer r.Close()
+			in, err := ioutil.ReadAll(r)
+			if err != nil {
+				b.Fatal(err)
+			}
+			got, err := zstd.Decompress(nil, in)
+			if err != nil {
+				b.Fatal(err)
+			}
+			b.SetBytes(int64(len(got)))
+			b.ReportAllocs()
+			b.ResetTimer()
+			b.RunParallel(func(pb *testing.PB) {
+				got := make([]byte, len(got))
+				for pb.Next() {
+					got, err = zstd.Decompress(got, in)
+					if err != nil {
+						b.Fatal(err)
+					}
+				}
+			})
+		})
+	}
+}
+
 func BenchmarkDecoderSilesiaCgo(b *testing.B) {
 	fn := "testdata/silesia.tar.zst"
 	data, err := ioutil.ReadFile(fn)
@@ -882,13 +1224,15 @@ func testDecoderDecodeAll(t *testing.T, fn string, dec *Decoder) {
 		}
 		want[tt.Name+".zst"], _ = ioutil.ReadAll(r)
 	}
-
-	for _, tt := range zr.File {
+	var wg sync.WaitGroup
+	for i, tt := range zr.File {
 		tt := tt
-		if !strings.HasSuffix(tt.Name, ".zst") {
+		if !strings.HasSuffix(tt.Name, ".zst") || (testing.Short() && i > 20) {
 			continue
 		}
+		wg.Add(1)
 		t.Run("DecodeAll-"+tt.Name, func(t *testing.T) {
+			defer wg.Done()
 			t.Parallel()
 			r, err := tt.Open()
 			if err != nil {
@@ -904,6 +1248,9 @@ func testDecoderDecodeAll(t *testing.T, fn string, dec *Decoder) {
 			if err != nil {
 				t.Error(err)
 			}
+			if len(got) < 10 {
+				t.Fatal("didn't get input back")
+			}
 			got = got[10:]
 			if !bytes.Equal(wantB, got) {
 				if len(wantB)+len(got) < 1000 {
@@ -926,6 +1273,10 @@ func testDecoderDecodeAll(t *testing.T, fn string, dec *Decoder) {
 			t.Log(len(got), "bytes returned, matches input, ok!")
 		})
 	}
+	go func() {
+		wg.Wait()
+		dec.Close()
+	}()
 }
 
 func testDecoderDecodeAllError(t *testing.T, fn string, dec *Decoder) {
@@ -938,12 +1289,15 @@ func testDecoderDecodeAllError(t *testing.T, fn string, dec *Decoder) {
 		t.Fatal(err)
 	}
 
+	var wg sync.WaitGroup
 	for _, tt := range zr.File {
 		tt := tt
 		if !strings.HasSuffix(tt.Name, ".zst") {
 			continue
 		}
+		wg.Add(1)
 		t.Run("DecodeAll-"+tt.Name, func(t *testing.T) {
+			defer wg.Done()
 			t.Parallel()
 			r, err := tt.Open()
 			if err != nil {
@@ -960,12 +1314,17 @@ func testDecoderDecodeAllError(t *testing.T, fn string, dec *Decoder) {
 			}
 		})
 	}
+	go func() {
+		wg.Wait()
+		dec.Close()
+	}()
 }
 
 // Test our predefined tables are correct.
 // We don't predefine them, since this also tests our transformations.
 // Reference from here: https://github.com/facebook/zstd/blob/ededcfca57366461021c922720878c81a5854a0a/lib/decompress/zstd_decompress_block.c#L234
 func TestPredefTables(t *testing.T) {
+	initPredefined()
 	x := func(nextState uint16, nbAddBits, nbBits uint8, baseVal uint32) decSymbol {
 		return newDecSymbol(nbBits, nbAddBits, nextState, baseVal)
 	}
@@ -1075,6 +1434,48 @@ func TestPredefTables(t *testing.T) {
 	}
 }
 
+func TestResetNil(t *testing.T) {
+	dec, err := NewReader(nil)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer dec.Close()
+
+	_, err = ioutil.ReadAll(dec)
+	if err != ErrDecoderNilInput {
+		t.Fatalf("Expected ErrDecoderNilInput when decoding from a nil reader, got %v", err)
+	}
+
+	emptyZstdBlob := []byte{40, 181, 47, 253, 32, 0, 1, 0, 0}
+
+	dec.Reset(bytes.NewBuffer(emptyZstdBlob))
+
+	result, err := ioutil.ReadAll(dec)
+	if err != nil && err != io.EOF {
+		t.Fatal(err)
+	}
+	if len(result) != 0 {
+		t.Fatalf("Expected to read 0 bytes, actually read %d", len(result))
+	}
+
+	dec.Reset(nil)
+
+	_, err = ioutil.ReadAll(dec)
+	if err != ErrDecoderNilInput {
+		t.Fatalf("Expected ErrDecoderNilInput when decoding from a nil reader, got %v", err)
+	}
+
+	dec.Reset(bytes.NewBuffer(emptyZstdBlob))
+
+	result, err = ioutil.ReadAll(dec)
+	if err != nil && err != io.EOF {
+		t.Fatal(err)
+	}
+	if len(result) != 0 {
+		t.Fatalf("Expected to read 0 bytes, actually read %d", len(result))
+	}
+}
+
 func timeout(after time.Duration) (cancel func()) {
 	c := time.After(after)
 	cc := make(chan struct{})
