@@ -37,7 +37,7 @@ func TestEncoder_EncodeAllSimple(t *testing.T) {
 	in = append(in, in...)
 	for level := EncoderLevel(speedNotSet + 1); level < speedLast; level++ {
 		t.Run(level.String(), func(t *testing.T) {
-			e, err := NewWriter(nil, WithEncoderLevel(level))
+			e, err := NewWriter(nil, WithEncoderLevel(level), WithEncoderConcurrency(2), WithWindowSize(128<<10), WithZeroFrames(true))
 			if err != nil {
 				t.Fatal(err)
 			}
@@ -211,7 +211,7 @@ func TestEncoderRegression(t *testing.T) {
 							encoded := enc.EncodeAll(in, nil)
 							got, err := dec.DecodeAll(encoded, nil)
 							if err != nil {
-								t.Logf("error: %v\nwant: %v\ngot:  %v", err, in, got)
+								t.Logf("error: %v\nwant: %v\ngot:  %v", err, len(in), len(got))
 								t.Fatal(err)
 							}
 
@@ -458,8 +458,10 @@ func TestEncoder_EncoderEnwik9(t *testing.T) {
 
 // test roundtrip using io.ReaderFrom interface.
 func testEncoderRoundtrip(t *testing.T, file string, wantCRC []byte) {
-	for level := EncoderLevel(speedNotSet + 1); level < speedLast; level++ {
+	for level := speedNotSet + 1; level < speedLast; level++ {
 		t.Run(level.String(), func(t *testing.T) {
+			level := level
+			t.Parallel()
 			f, err := os.Open(file)
 			if err != nil {
 				if os.IsNotExist(err) {
@@ -667,6 +669,76 @@ func TestEncoder_EncodeAllSilesia(t *testing.T) {
 	t.Log("Encoded content matched")
 }
 
+func TestEncoderReadFrom(t *testing.T) {
+	buffer := bytes.NewBuffer(nil)
+	encoder, err := NewWriter(buffer)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if _, err := encoder.ReadFrom(strings.NewReader("0")); err != nil {
+		t.Fatal(err)
+	}
+	if err := encoder.Close(); err != nil {
+		t.Fatal(err)
+	}
+
+	dec, _ := NewReader(nil)
+	toDec := buffer.Bytes()
+	toDec = append(toDec, toDec...)
+	decoded, err := dec.DecodeAll(toDec, nil)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if !bytes.Equal([]byte("00"), decoded) {
+		t.Logf("encoded: % x\n", buffer.Bytes())
+		t.Fatalf("output mismatch, got %s", string(decoded))
+	}
+	dec.Close()
+}
+
+func TestInterleavedWriteReadFrom(t *testing.T) {
+	var encoded bytes.Buffer
+
+	enc, err := NewWriter(&encoded)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if _, err := enc.Write([]byte("write1")); err != nil {
+		t.Fatal(err)
+	}
+	if _, err := enc.Write([]byte("write2")); err != nil {
+		t.Fatal(err)
+	}
+	if _, err := enc.ReadFrom(strings.NewReader("readfrom1")); err != nil {
+		t.Fatal(err)
+	}
+	if _, err := enc.Write([]byte("write3")); err != nil {
+		t.Fatal(err)
+	}
+
+	if err := enc.Close(); err != nil {
+		t.Fatal(err)
+	}
+
+	dec, err := NewReader(&encoded)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer dec.Close()
+
+	gotb, err := ioutil.ReadAll(dec)
+	if err != nil {
+		t.Fatal(err)
+	}
+	got := string(gotb)
+
+	if want := "write1write2readfrom1write3"; got != want {
+		t.Errorf("got decoded %q, want %q", got, want)
+	}
+}
+
 func TestEncoder_EncodeAllEmpty(t *testing.T) {
 	if testing.Short() {
 		t.SkipNow()
@@ -927,9 +999,9 @@ func BenchmarkEncoder_EncodeAllPi(b *testing.B) {
 	}
 }
 
-func BenchmarkRandomEncodeAllFastest(b *testing.B) {
+func BenchmarkRandom4KEncodeAllFastest(b *testing.B) {
 	rng := rand.New(rand.NewSource(1))
-	data := make([]byte, 10<<20)
+	data := make([]byte, 4<<10)
 	for i := range data {
 		data[i] = uint8(rng.Intn(256))
 	}
@@ -948,12 +1020,29 @@ func BenchmarkRandomEncodeAllFastest(b *testing.B) {
 	}
 }
 
-func BenchmarkRandomEncodeAllDefault(b *testing.B) {
+func BenchmarkRandom10MBEncodeAllFastest(b *testing.B) {
 	rng := rand.New(rand.NewSource(1))
 	data := make([]byte, 10<<20)
-	for i := range data {
-		data[i] = uint8(rng.Intn(256))
+	rng.Read(data)
+	enc, _ := NewWriter(nil, WithEncoderLevel(SpeedFastest), WithEncoderConcurrency(1))
+	defer enc.Close()
+	dst := enc.EncodeAll(data, nil)
+	wantSize := len(dst)
+	b.ResetTimer()
+	b.ReportAllocs()
+	b.SetBytes(int64(len(data)))
+	for i := 0; i < b.N; i++ {
+		dst := enc.EncodeAll(data, dst[:0])
+		if len(dst) != wantSize {
+			b.Fatal(len(dst), "!=", wantSize)
+		}
 	}
+}
+
+func BenchmarkRandom4KEncodeAllDefault(b *testing.B) {
+	rng := rand.New(rand.NewSource(1))
+	data := make([]byte, 4<<10)
+	rng.Read(data)
 	enc, _ := NewWriter(nil, WithEncoderLevel(SpeedDefault), WithEncoderConcurrency(1))
 	defer enc.Close()
 	dst := enc.EncodeAll(data, nil)
@@ -969,12 +1058,29 @@ func BenchmarkRandomEncodeAllDefault(b *testing.B) {
 	}
 }
 
-func BenchmarkRandomEncoderFastest(b *testing.B) {
+func BenchmarkRandomEncodeAllDefault(b *testing.B) {
 	rng := rand.New(rand.NewSource(1))
 	data := make([]byte, 10<<20)
-	for i := range data {
-		data[i] = uint8(rng.Intn(256))
+	rng.Read(data)
+	enc, _ := NewWriter(nil, WithEncoderLevel(SpeedDefault), WithEncoderConcurrency(1))
+	defer enc.Close()
+	dst := enc.EncodeAll(data, nil)
+	wantSize := len(dst)
+	b.ResetTimer()
+	b.ReportAllocs()
+	b.SetBytes(int64(len(data)))
+	for i := 0; i < b.N; i++ {
+		dst := enc.EncodeAll(data, dst[:0])
+		if len(dst) != wantSize {
+			b.Fatal(len(dst), "!=", wantSize)
+		}
 	}
+}
+
+func BenchmarkRandom10MBEncoderFastest(b *testing.B) {
+	rng := rand.New(rand.NewSource(1))
+	data := make([]byte, 10<<20)
+	rng.Read(data)
 	wantSize := int64(len(data))
 	enc, _ := NewWriter(ioutil.Discard, WithEncoderLevel(SpeedFastest))
 	defer enc.Close()
@@ -1003,9 +1109,7 @@ func BenchmarkRandomEncoderFastest(b *testing.B) {
 func BenchmarkRandomEncoderDefault(b *testing.B) {
 	rng := rand.New(rand.NewSource(1))
 	data := make([]byte, 10<<20)
-	for i := range data {
-		data[i] = uint8(rng.Intn(256))
-	}
+	rng.Read(data)
 	wantSize := int64(len(data))
 	enc, _ := NewWriter(ioutil.Discard, WithEncoderLevel(SpeedDefault))
 	defer enc.Close()
